use std::{str::FromStr, sync::Arc};

use derive_more::derive::{Display, Into};
use serde::{Deserialize, Serialize};
use thiserror::Error;

use crate::constants;

#[derive(Debug, Error, PartialEq, Eq)]
pub enum InvalidGivenDayError {
    #[error(
        "({0}) is not a valid format for unsigned whole number between 0 and {}",
        constants::MAX_DAY
    )]
    InvalidFormat(Arc<str>),
    #[error("({0}) is within the valid range between 0 and {}", constants::MAX_DAY)]
    InvalidRange(u32),
}

#[derive(Debug, Copy, Clone, Into, PartialEq, Eq, Serialize, Deserialize, Display)]
pub struct GivenDay(u32);

impl GivenDay {
    pub fn new(value: u32) -> Result<Self, InvalidGivenDayError> {
        if value < 1 || value > constants::MAX_DAY {
            Err(InvalidGivenDayError::InvalidRange(value))
        } else {
            Ok(Self(value))
        }
    }
}
impl FromStr for GivenDay {
    type Err = InvalidGivenDayError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let number: u32 = s
            .parse()
            .map_err(|_| InvalidGivenDayError::InvalidFormat(Arc::from(s)))?;

        GivenDay::new(number)
    }
}

#[cfg(test)]
mod testing {
    use crate::constants::{self, MAX_DAY};

    use super::*;

    #[test]
    fn parse_correct_value() {
        fn assert_case(input: &str, expected: u32) {
            let actual = GivenDay::from_str(input).unwrap();
            assert_eq!(expected, actual.into(), "input: {}", input);
        }

        assert_case("1", 1);
        assert_case("3", 3);
        assert_case(&constants::MAX_DAY.to_string(), MAX_DAY);
    }

    #[test]
    fn detect_invalid_input() {
        fn assert_case(input: &str, expected_err: InvalidGivenDayError) {
            let actual = GivenDay::from_str(input).unwrap_err();
            assert_eq!(expected_err, actual, "input: {}", input);
        }

        const OVER_MAX: u32 = constants::MAX_DAY + 1;
        assert_case("0", InvalidGivenDayError::InvalidRange(0));
        assert_case(
            &OVER_MAX.to_string(),
            InvalidGivenDayError::InvalidRange(OVER_MAX),
        );
        assert_case("", InvalidGivenDayError::InvalidFormat("".into()));
        assert_case("a", InvalidGivenDayError::InvalidFormat("a".into()));
    }
}
